/* OpenVAS Administrator
 * $Id$
 * Description: Module for Administrator: the Admin library.
 *
 * Authors:
 * Matthew Mundell <matt@mundell.ukfsn.org>
 * Michael Wiegand <michael.wiegand@intevation.de>
 *
 * Copyright:
 * Copyright (C) 2009 Greenbone Networks GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * or, at your option, any later version as published by the Free
 * Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * @file  admin.c
 * @brief The OpenVAS administration library.
 *
 * This file defines an administration library, for building OpenVAS server
 * administration clients.
 */

#include "admin.h"

#include <assert.h>
#include <gcrypt.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <openvas/openvas_auth.h>

#ifdef S_SPLINT_S
#include "splint.h"
#endif

#undef G_LOG_DOMAIN
/**
 * @brief GLib log domain.
 */
#define G_LOG_DOMAIN "ad  admin"

/** @todo Add contention handling at this level. */

/**
 * @brief Convenience function to produce text output from key/value pairs of
 * settings.
 * This function is used by g_hash_table_foreach().
 *
 * @param key The key.
 * @param value The value corresponding to the key.
 * @param user_data A GString object to which the output should be appended.
 */
static void
print_setting (void * key, void * value, void * user_data)
{
  gchar *text;
  text = g_strdup_printf ("%s=%s\n", (gchar *) key, (gchar *) value);
  g_string_append ((GString *) user_data, text);
  g_free (text);
}

/**
 * @brief Convenience function to produce XML output from user list items.
 * This function is used by g_slist_foreach().
 *
 * @param data The item's data.
 * @param user_data User data passed to g_slist_foreach().
 */
static void
print_user (void * data, void * user_data)
{
  gchar *xml;
  xml = g_markup_printf_escaped ("<user><name>%s</name><rules/></user>",
                                 (gchar *) data);
  g_string_append ((GString *) user_data, xml);
  g_free (xml);
}

/**
 * @brief Convenience function to produce text output from user list items.
 * This function is used by g_slist_foreach().
 *
 * @param data The item's data.
 * @param user_data User data passed to g_slist_foreach().
 */
static void
print_user_text (void * data, void * user_data)
{
  gchar *xml;
  xml = g_markup_printf_escaped ("%s\n", (gchar *) data);
  g_string_append ((GString *) user_data, xml);
  g_free (xml);
}

/**
 * @brief Checks whether a file is a directory or not.
 *
 * This is a replacement for the g_file_test functionality which is reported
 * to be unreliable under certain circumstances, for example if this
 * application and glib are compiled with a different libc.
 *
 * @todo FIXME: handle symbolic links
 * @todo Move to libs?
 *
 * @return 1 if parameter is directory, 0 if it is not, -1 if it does not
 * exist or could not be accessed.
 */
static int
check_is_dir (const char* name)
{
  struct stat sb;

  if (stat (name, &sb))
    {
      return -1;
    }
  else
    {
      return (S_ISDIR (sb.st_mode));
    }
}

/**
 * @brief Recursively removes files and directories.
 *
 * This function will recursively call itself to delete a path and any
 * contents of this path.
 *
 * @todo Move to libs?
 *
 * @param pathname The name of the file to be deleted from the filesystem.
 *
 * @return 0 if the name was successfully deleted, -1 if an error occurred.
 * Please note that errno is currently not guaranteed to contain the correct
 * value if -1 is returned.
 */
static int
remove_recurse (const gchar * pathname)
{
  if (check_is_dir (pathname) == 1)
    {
      GError *error = NULL;
      GDir *directory = g_dir_open (pathname, 0, &error);

      if (directory == NULL)
        {
          g_warning ("g_dir_open(%s) failed - %s\n", pathname, error->message);
          g_error_free (error);
          // errno should be set when we return -1 to maintain remove()
          // compatibility.
          return -1;
        }
      else
        {
          int ret = 0;
          const gchar *entry = NULL;

          while ((entry = g_dir_read_name (directory)) && (ret == 0))
            {
              gchar *entry_path = g_build_filename (pathname, entry, NULL);
              ret = remove_recurse (entry_path);
              g_free (entry_path);
              if (ret != 0)
                {
                  g_warning ("Failed to remove %s from %s!", entry, pathname);
                  g_dir_close (directory);
                  return ret;
                }
            }
          g_dir_close (directory);
        }
    }

  return g_remove (pathname);
}

/**
 * @brief Returns a list of user directories (= users) found in a given
 * directory.
 *
 * @param directory The complete name of the directory.
 *
 * @return A pointer to a GSList containing the names of the users or NULL if
 * the directory could not be opened, did not exist or was not a directory.
 * The list should be freed with g_slist_free when no longer needed.  Each
 * element of the list should be freed with g_free.
 */
GSList *
openvas_admin_list_users (const gchar * directory)
{
  GSList *users = NULL;

  if (g_file_test (directory, G_FILE_TEST_EXISTS) &&
      g_file_test (directory, G_FILE_TEST_IS_DIR))
    {
      const gchar *entry_name = NULL;
      GError *error = NULL;
      GDir *users_dir = NULL;

      users_dir = g_dir_open (directory, 0, &error);
      if (users_dir == NULL)
        {
          g_warning (error->message);
          g_error_free (error);
        }
      else
        {
          while ((entry_name = g_dir_read_name (users_dir)))
            {
              gchar *user_hash_filename;
              gchar *user_dname_filename;

              if (strcmp (entry_name, "om") == 0) continue;

              user_hash_filename = g_build_filename (directory, entry_name,
                                                     "auth", "hash", NULL);
              user_dname_filename = g_build_filename (directory, entry_name,
                                                      "auth", "dname", NULL);
              if (g_file_test (user_hash_filename, G_FILE_TEST_EXISTS))
                {
                  users = g_slist_append (users, g_strdup (entry_name));
                }
              else if (g_file_test (user_dname_filename, G_FILE_TEST_EXISTS))
                {
                  users = g_slist_append (users, g_strdup (entry_name));
                }
              g_free (user_hash_filename);
              g_free (user_dname_filename);
            }
          g_dir_close (users_dir);
        }

      return users;
    }
  else
    {
      g_warning ("Could not find %s!", directory);
      return NULL;
    }
}

/** @todo Get the 200's in these commands from libs (STATUS_OK). */

/**
 * @brief Produces a newline separated list from a GHashTable of settings.
 *
 * @param preferences A pointer to a GHashTable containing settings.
 *
 * @return A pointer to a GString containing the settings list.
 */
GString *
print_settings (GHashTable * settings)
{
  GString *response = g_string_new ("");
  if (settings)
    {
      g_hash_table_foreach (settings, print_setting, response);
    }
  return response;
}

/**
 * @brief Produces an ocp_response element for a list_users request from a
 * GSList of users.
 *
 * @param users A pointer to a GSList containing user names.
 *
 * @return A pointer to a GString containig the ocp_response element.
 */
GString *
print_users_xml (GSList * users)
{
  GString *response = NULL;
  if (users)
    {
      response = g_string_new ("<get_users_reponse status=\"200\">");
      g_slist_foreach (users, print_user, response);
      response = g_string_append (response, "</get_users_response>");
    }
  else
    {
      response = g_string_new ("<get_users_response status=\"200\">"
                               "<users/>"
                               "</get_users_response>");
    }
  return response;
}

/**
 * @brief Produces a simple newline separated list of user names from a
 * GSList of users.
 *
 * @param users A pointer to a GSList containing user names.
 *
 * @return A pointer to a GString containig the ocp_response element.
 */
GString *
print_users_text (GSList * users)
{
  GString *response = g_string_new("");
  if (users)
    {
      g_slist_foreach (users, print_user_text, response);
    }
  return response;
}

/**
 * @brief Adds a new user to the OpenVAS installation.
 *
 * @param name The name of the new user.
 * @param password The password of the new user.
 * @param role The role of the user.
 * @param directory The directory containing the user directories. It will be
 * created if it does not exist already.
 *
 * \todo TODO: Adding users authenticating with certificates is not yet
 * implemented.
 *
 * @return 0 if the user has been added successfully, -1 on error, -2 if user
 *         exists already.
 */
int
openvas_admin_add_user (const gchar * name, const gchar * password,
                        const gchar * role, const gchar * directory)
{
  const gchar *rules = "";

  assert (name != NULL);
  assert (password != NULL);
  assert (role != NULL);
  assert (directory != NULL);

  if (strcmp (name, "om") == 0)
    {
      g_warning ("Attempt to add special \"om\" user!");
      return -1;
    }

  if (! g_file_test (directory, G_FILE_TEST_EXISTS))
    {
      if (g_mkdir (directory, 0700) != 0)
        {
          g_warning ("Could not create %s!", directory);
          return -1;
        }
    }

  if (g_file_test (directory, G_FILE_TEST_IS_DIR))
    {
      gchar *user_dir_name = g_build_filename (directory, name, NULL);

      if (g_file_test (user_dir_name, G_FILE_TEST_EXISTS) &&
          g_file_test (user_dir_name, G_FILE_TEST_IS_DIR))
        {
          g_warning ("User %s already exists!", name);
          g_free (user_dir_name);
          return -2;
        }
      else
        {
          GError *error = NULL;
          gchar *hashes_out;

          gchar *user_auth_dir_name = g_build_filename (user_dir_name, "auth", NULL);
          gchar *user_rules_file_name = g_build_filename (user_auth_dir_name, "rules", NULL);
          gchar *user_hash_file_name = g_build_filename (user_auth_dir_name, "hash", NULL);
          gboolean directories_created = TRUE;

          if (g_mkdir (user_dir_name, 0700) != 0)
            {
              g_warning ("Could not create %s!", user_dir_name);
              directories_created =  -1;
            }
          else if (g_mkdir (user_auth_dir_name, 0700) != 0)
            {
              g_warning ("Could not create %s!", user_auth_dir_name);
              if (remove_recurse (user_dir_name) != 0)
                g_warning ("Could not remove %s while trying to revert changes!", user_dir_name);
              directories_created =  -1;
            }

          if (!directories_created)
            {
              g_free (user_dir_name);
              g_free (user_auth_dir_name);
              g_free (user_rules_file_name);
              g_free (user_hash_file_name);
              g_warning ("Failed to set up user directories for user %s", name);
              return -1;
            }

          hashes_out = get_password_hashes (GCRY_MD_MD5, password);

          if (!g_file_set_contents (user_hash_file_name, hashes_out, -1, &error))
            {
              g_warning ("%s", error->message);
              g_error_free (error);
              if (remove_recurse (user_dir_name) != 0)
                g_warning ("Could not remove %s while trying to revert changes!", user_dir_name);
              g_free (hashes_out);
              g_free (user_dir_name);
              g_free (user_auth_dir_name);
              g_free (user_rules_file_name);
              g_free (user_hash_file_name);
              return -1;
            }
          g_chmod (user_hash_file_name, 0600);

          if (!g_file_set_contents (user_rules_file_name,
                                    rules ? rules : "",
                                    -1,
                                    &error))
            {
              g_warning ("%s", error->message);
              g_error_free (error);
              if (remove_recurse (user_dir_name) != 0)
                g_warning ("Could not remove %s while trying to revert changes!", user_dir_name);
              g_free (hashes_out);
              g_free (user_dir_name);
              g_free (user_auth_dir_name);
              g_free (user_rules_file_name);
              g_free (user_hash_file_name);
              return -1;
            }
          g_chmod (user_rules_file_name, 0600);

          if (openvas_set_user_role (name, role))
            {
              if (remove_recurse (user_dir_name) != 0)
                g_warning ("Could not remove %s while trying to revert changes!",
                           user_dir_name);
              g_free (hashes_out);
              g_free (user_dir_name);
              g_free (user_auth_dir_name);
              g_free (user_rules_file_name);
              g_free (user_hash_file_name);
              return -1;
            }

          g_free (hashes_out);
          g_free (user_dir_name);
          g_free (user_auth_dir_name);
          g_free (user_rules_file_name);
          g_free (user_hash_file_name);

          return 0;
        }
    }
  else
    {
      g_warning ("Could not access %s!", directory);
      return -1;
    }
}

/**
 * @brief Removes an user from the OpenVAS installation.
 *
 * @param name The name of the user to be removed.
 * @param directory The directory containing the user directories.
 *
 * @return 0 if the user has been removed successfully, -1 on error,
 *         -2 if failed to find such a user.
 */
int
openvas_admin_remove_user (const gchar * name, const gchar * directory)
{
  assert (name != NULL);

  if (strcmp (name, "om") == 0)
    {
      g_warning ("Attempt to remove special \"om\" user!");
      return -1;
    }

  if (g_file_test (directory, G_FILE_TEST_EXISTS) &&
      g_file_test (directory, G_FILE_TEST_IS_DIR))
    {
      gchar *user_dir_name = g_build_filename (directory, name, NULL);

      if (g_file_test (user_dir_name, G_FILE_TEST_EXISTS) &&
          g_file_test (user_dir_name, G_FILE_TEST_IS_DIR))
        {
          if (remove_recurse (user_dir_name) == 0)
            {
              g_free (user_dir_name);
              return 0;
            }
          else
            {
              g_warning ("Failed to remove %s!", user_dir_name);
              g_free (user_dir_name);
              return -1;
            }
        }
      else
        {
          g_free (user_dir_name);
          g_warning ("User %s does not exist!", name);
          return -2;
        }
    }
  else
    {
      g_warning ("Could not find %s!", directory);
      return -1;
    }
}

/**
 * @brief Sets the rules for an existing user.
 *
 * @param name The name of the user.
 * @param rules_file A file containing the new rules to be applied to the
 * user. Note that the new rules will overwrite all old rules.
 * @param directory The directory containing the user directories.
 *
 * @return TRUE if the rules have been changed successfully, FALSE if not.
 */
gboolean
openvas_admin_set_rules (const gchar * name, const gchar * rules_file,
                         const gchar * directory)
{
  if (strcmp (name, "om") == 0)
    {
      g_warning ("Attempt to set rules of special \"om\" user!");
      return FALSE;
    }
  if (name == NULL || rules_file == NULL)
    {
      g_warning ("Missing name or rules_file!");
      return FALSE;
    }
  if (g_file_test (directory, G_FILE_TEST_EXISTS) &&
      g_file_test (directory, G_FILE_TEST_IS_DIR))
    {
      gchar *user_dir_name = g_build_filename (directory, name, NULL);

      if (g_file_test (user_dir_name, G_FILE_TEST_EXISTS) &&
          g_file_test (user_dir_name, G_FILE_TEST_IS_DIR))
        {
          gchar *new_rules = NULL;
          GError *error = NULL;
          gchar *user_rules_file_name = NULL;

          if (!g_file_test (rules_file, G_FILE_TEST_EXISTS))
            {
              g_warning ("Could not find rules file %s!", rules_file);
              g_free (user_dir_name);
              return FALSE;
            }
          if (!g_file_get_contents (rules_file, &new_rules, NULL, &error))
            {
              g_warning ("Could not read contents of %s!", rules_file);
              g_warning ("%s", error->message);
              g_error_free (error);
              g_free (user_dir_name);
              return FALSE;
            }

          user_rules_file_name = g_build_filename (user_dir_name,
                                                   "auth",
                                                   "rules",
                                                   NULL);

          if (!g_file_set_contents (user_rules_file_name,
                                    new_rules,
                                    -1,
                                    &error))
            {
              g_warning ("%s", error->message);
              g_error_free (error);
              g_free (user_dir_name);
              g_free (user_rules_file_name);
              return FALSE;
            }
          g_chmod (user_rules_file_name, 0600);

          g_free (user_rules_file_name);
          g_free (user_dir_name);
          return TRUE;
        }
      else
        {
          g_warning ("User %s does not exist!", name);
          g_free (user_dir_name);
          return FALSE;
        }
    }
  else
    {
      g_warning ("Could not find %s!", directory);
      return FALSE;
    }
}

/**
 * @brief Ask an NVT synchronization script to perform a selftest and report
 * the results.
 *
 * @param sync_script The file name of the synchronization script.
 * @param [out] Return location for selftest errors, or NULL.
 *
 * @return TRUE if the selftest was successful, or FALSE if an error occured.
 */
gboolean
openvas_sync_script_perform_selftest (const gchar * sync_script,
                                      gchar **result)
{
  g_assert (sync_script);
  g_assert_cmpstr (*result, ==, NULL);

  gchar *script_working_dir = g_path_get_dirname (sync_script);

  gchar ** argv = (gchar **) g_malloc (3 * sizeof (gchar *));
  argv[0] = g_strdup (sync_script);
  argv[1] = g_strdup ("--selftest");
  argv[2] = NULL;

  gchar *script_out;
  gchar *script_err;
  gint script_exit;
  GError *error = NULL;

  if (!g_spawn_sync (script_working_dir, argv, NULL, 0, NULL, NULL,
                     &script_out, &script_err, &script_exit, &error))
    {
      if (*result != NULL)
        {
          *result = g_strdup_printf ("Failed to execute synchronization "
                                     "script: %s", error->message);
        }

      g_free (script_working_dir);
      g_strfreev (argv);
      g_free (script_out);
      g_free (script_err);
      g_free (error);

      return FALSE;
    }

  if (script_exit != 0)
    {
      if (*result != NULL)
        {
          *result = g_strdup_printf ("%s", script_err);
        }

      g_free (script_working_dir);
      g_strfreev (argv);
      g_free (script_out);
      g_free (script_err);
      g_free (error);

      return FALSE;
    }

  g_free (script_working_dir);
  g_strfreev (argv);
  g_free (script_out);
  g_free (script_err);
  g_free (error);

  return TRUE;
}

/**
 * @brief Retrieves the identification string of an NVT synchronization
 * script, with basic validation.
 *
 * @param sync_script The file name of the synchronization script.
 * @param [out] Return location of the identification string.
 *
 * @return TRUE if the identification string was retrieved, or FALSE if an
 * error occured.
 */
gboolean
openvas_get_sync_script_identification (const gchar * sync_script,
                                        gchar **identification)
{
  g_assert (sync_script);
  g_assert_cmpstr (*identification, ==, NULL);

  gchar *script_working_dir = g_path_get_dirname (sync_script);

  gchar ** argv = (gchar **) g_malloc (3 * sizeof (gchar *));
  argv[0] = g_strdup (sync_script);
  argv[1] = g_strdup ("--identify");
  argv[2] = NULL;

  gchar *script_out;
  gchar *script_err;
  gint script_exit;
  GError *error = NULL;

  gchar **script_identification;

  if (!g_spawn_sync (script_working_dir, argv, NULL, 0, NULL, NULL,
                     &script_out, &script_err, &script_exit, &error))
    {
      g_warning ("Failed to execute %s: %s", sync_script, error->message);

      g_free (script_working_dir);
      g_strfreev (argv);
      g_free (script_out);
      g_free (script_err);
      g_free (error);

      return FALSE;
    }

  if (script_exit != 0)
    {
      g_warning ("%s returned a non-zero exit code.", sync_script);

      g_free (script_working_dir);
      g_strfreev (argv);
      g_free (script_out);
      g_free (script_err);
      g_free (error);

      return FALSE;
    }

  script_identification = g_strsplit (script_out, "|", 6);

  if (g_strncasecmp (script_identification[0], "NVTSYNC", 7) ||
      g_strncasecmp (script_identification[0], script_identification[5], 7))
    {
      g_warning ("%s is not a NVT synchronization script.", sync_script);

      g_free (script_working_dir);
      g_strfreev (argv);
      g_free (script_out);
      g_free (script_err);
      g_free (error);

      g_strfreev (script_identification);

      return FALSE;
    }

  *identification = g_strdup (script_out);

  g_free (script_working_dir);
  g_strfreev (argv);
  g_free (script_out);
  g_free (script_err);
  g_free (error);

  g_strfreev (script_identification);

  return TRUE;
}

/**
 * @brief Retrieves the description of an NVT synchronization
 * script, with basic validation.
 *
 * @param sync_script The file name of the synchronization script.
 * @param [out] Return location of the description string.
 *
 * @return TRUE if the description was retrieved, or FALSE if an error
 * occured.
 */
gboolean
openvas_get_sync_script_description (const gchar * sync_script,
                                     gchar **description)
{
  g_assert (sync_script);
  g_assert_cmpstr (*description, ==, NULL);

  gchar *script_working_dir = g_path_get_dirname (sync_script);

  gchar ** argv = (gchar **) g_malloc (3 * sizeof (gchar *));
  argv[0] = g_strdup (sync_script);
  argv[1] = g_strdup ("--describe");
  argv[2] = NULL;

  gchar *script_out;
  gchar *script_err;
  gint script_exit;
  GError *error = NULL;

  if (!g_spawn_sync (script_working_dir, argv, NULL, 0, NULL, NULL,
                     &script_out, &script_err, &script_exit, &error))
    {
      g_warning ("Failed to execute %s: %s", sync_script, error->message);

      g_free (script_working_dir);
      g_strfreev (argv);
      g_free (script_out);
      g_free (script_err);
      g_free (error);

      return FALSE;
    }

  if (script_exit != 0)
    {
      g_warning ("%s returned a non-zero exit code.", sync_script);

      g_free (script_working_dir);
      g_strfreev (argv);
      g_free (script_out);
      g_free (script_err);
      g_free (error);

      return FALSE;
    }

  *description = g_strdup (script_out);

  g_free (script_working_dir);
  g_strfreev (argv);
  g_free (script_out);
  g_free (script_err);
  g_free (error);

  return TRUE;
}

/**
 * @brief Calls a NVT synchronization script to synchronize the local NVT
 * collection.
 *
 * @param sync_script The file name of the synchronization script.
 *
 * @return TRUE if the synchronization has finished successfully, FALSE if
 * not.
 */
gboolean
openvas_sync_feed (const gchar * sync_script)
{
  g_assert (sync_script);

  gchar *script_identification_string = NULL;

  if (!openvas_get_sync_script_identification (sync_script,
                                               &script_identification_string))
    {
      g_warning ("No valid synchronization script supplied!");
      return FALSE;
    }

  gchar *script_working_dir = g_path_get_dirname (sync_script);

  gchar ** argv = (gchar **) g_malloc (2 * sizeof (gchar *));
  argv[0] = g_strdup (sync_script);
  argv[1] = NULL;

  gchar *script_out;
  gchar *script_err;
  gint script_exit;
  GError *error = NULL;

  if (!g_spawn_sync (script_working_dir, argv, NULL, 0, NULL, NULL,
                     &script_out, &script_err, &script_exit, &error))
    {
      g_warning ("Failed to execute %s: %s", sync_script, error->message);

      g_free (script_working_dir);
      g_strfreev (argv);
      g_free (script_out);
      g_free (script_err);
      g_free (error);

      return FALSE;
    }

  if (script_exit != 0)
    {
      g_warning ("%s returned a non-zero exit code.", sync_script);

      g_free (script_working_dir);
      g_strfreev (argv);
      g_free (script_out);
      g_free (script_err);
      g_free (error);

      return FALSE;
    }

  g_free (script_working_dir);
  g_strfreev (argv);
  g_free (script_out);
  g_free (script_err);
  g_free (error);

  return TRUE;
}

/**
 * @brief Create a lockfile for synchronization and store the name of the
 * synchronization script.
 *
 * @param sync_script The file name of the synchronization script.
 *
 * @return TRUE if the lockfile was written successfully, FALSE if
 * not.
 */
gboolean
openvasad_create_lockfile (const gchar * sync_script)
{
  g_assert (sync_script);

  if (openvasad_currently_syncing ())
    {
      return FALSE;
    }

  gchar *lockfile_name = g_build_filename (g_get_tmp_dir (),
                                           "openvasad-feed-sync.lock",
                                           NULL);

  if (!g_file_set_contents (lockfile_name, sync_script, -1, NULL))
    {
      g_warning ("Failed to write lockfile!");
      g_free (lockfile_name);
      return FALSE;
    }

  g_free (lockfile_name);
  return TRUE;
}

/**
 * @brief Remove a lockfile for synchronization.
 *
 * @return TRUE if the lockfile was removed successfully, FALSE if
 * not.
 */
gboolean
openvasad_remove_lockfile (void)
{
  gchar *lockfile_name = g_build_filename (g_get_tmp_dir (),
                                           "openvasad-feed-sync.lock",
                                           NULL);
  if (!g_file_test (lockfile_name, G_FILE_TEST_EXISTS))
    {
      g_free (lockfile_name);
      return TRUE;
    }

  if (g_unlink (lockfile_name) == 0)
    {
      g_free (lockfile_name);
      return TRUE;
    }

  g_warning ("Failed to remove lockfile!");
  g_free (lockfile_name);
  return FALSE;
}

/**
 * @brief Determine if the administrator is currently synchronizing with a
 * feed, i.e. if a lockfile exists.
 *
 * @return TRUE if the administrator is currently synchronizing with a feed,
 * FALSE if not.
 */
gboolean
openvasad_currently_syncing (void)
{
  gchar *lockfile_name = g_build_filename (g_get_tmp_dir (),
                                           "openvasad-feed-sync.lock",
                                           NULL);
  if (g_file_test (lockfile_name, G_FILE_TEST_EXISTS))
    {
      g_free (lockfile_name);
      return TRUE;
    }

  g_free (lockfile_name);
  return FALSE;
}
