/* OpenVAS Administrator
 * $Id$
 * Description: Module for OpenVAS Administrator: the OAP library.
 *
 * Authors:
 * Matthew Mundell <matt@mundell.ukfsn.org>
 *
 * Copyright:
 * Copyright (C) 2009 Greenbone Networks GmbH
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2,
 * or, at your option, any later version as published by the Free
 * Software Foundation
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
 * @file  oap.c
 * @brief The OpenVAS Administrator OAP library.
 *
 * This file defines an OpenVAS Administrator Protocol (OAP) library, for
 * implementing OpenVAS administrators such as the OpenVAS Administrator daemon.
 *
 * The library provides \ref process_oap_client_input.
 * This function parses a given string of OAP XML and performs administration
 * tasks in reaction to the OAP commands in the string.
 */

#include "admin.h"
#include "oap.h"
#include "tracef.h"

#include <assert.h>
#include <string.h>

#include <openvas/base/openvas_string.h> /* for openvas_append_text */
#include <openvas/base/settings.h>
#include <openvas/openvas_auth.h>
#include <openvas/openvas_logging.h>

#ifdef S_SPLINT_S
#include "splint.h"
#endif

#undef G_LOG_DOMAIN
/**
 * @brief GLib log domain.
 */
#define G_LOG_DOMAIN "ad    oap"

/**
 * @brief The users directory for this daemon.
 */
static const gchar *users_dir = NULL;

/**
 * @brief The synchronization script for this daemon.
 */
static const gchar *sync_script = NULL;

/**
 * @brief The scanner configuration file for this daemon.
 */
static const gchar *scanner_config_file = NULL;


/* Credentials. */

/** @todo Same as in openvas-manager/src/manage.h. */

/**
 * @brief A username password pair.
 */
typedef struct
{
  /*@null@*/ gchar* username; ///< Login name of user.
  /*@null@*/ gchar* password; ///< Password of user.
} credentials_t;

/** @todo Copy of section from openvas-manager/src/manage.c. */

/**
 * @brief Current credentials during any OMP command.
 */
credentials_t current_credentials;

/**
 * @brief Free credentials.
 *
 * Free the members of a credentials pair.
 *
 * @param[in]  credentials  Pointer to the credentials.
 */
void
free_credentials (credentials_t* credentials)
{
  if (credentials->username)
    {
      g_free (credentials->username);
      credentials->username = NULL;
    }
  if (credentials->password)
    {
      g_free (credentials->password);
      credentials->password = NULL;
    }
}

/**
 * @brief Append text to the username of a credential pair.
 *
 * @param[in]  credentials  Credentials.
 * @param[in]  text         The text to append.
 * @param[in]  length       Length of the text.
 */
void
append_to_credentials_username (credentials_t* credentials,
                                const char* text,
                                gsize length)
{
  openvas_append_text (&credentials->username, text, length);
}

/**
 * @brief Append text to the password of a credential pair.
 *
 * @param[in]  credentials  Credentials.
 * @param[in]  text         The text to append.
 * @param[in]  length       Length of the text.
 */
void
append_to_credentials_password (credentials_t* credentials,
                                const char* text,
                                gsize length)
{
  openvas_append_text (&credentials->password, text, length);
}


/* Convenience functions. */

/**
 * @brief Convenience function to produce XML output from key/value pairs of
 * settings.
 * This function is used by g_hash_table_foreach().
 *
 * @param key The key.
 * @param value The value corresponding to the key.
 * @param xml_string GString where the resulting XML will be written.
 */
static void
build_settings_xml (void * key, void * value, void * xml_string)
{
  gchar *xml;
  xml = g_markup_printf_escaped ("<setting name=\"%s\">%s</setting>",
                                 (gchar *) key, (gchar *) value);
  g_string_append ((GString *) xml_string, xml);
  g_free (xml);
}


/* Authenticate. */

/**
 * @brief Authenticate credentials.
 *
 * @param[in]  credentials  Credentials.
 *
 * @return 0 authentication success, 1 authentication failure, -1 error.
 */
int
authenticate (credentials_t* credentials)
{
  if (credentials->username
      && credentials->password
      && openvas_is_user_admin (credentials->username))
    {
      int fail;

      if (strcmp (credentials->username, "om") == 0) return 1;

      fail = openvas_authenticate (credentials->username,
                                   credentials->password);
      if (fail == 0)
        {
          return 0;
        }
      return fail;
    }
  return 1;
}


/* Help message. */

static char* help_text = "\n"
"    AUTHENTICATE           Authenticate with the administrator.\n"
"    COMMANDS               Run a list of commands.\n"
"    CREATE_USER            Create a new user.\n"
"    DELETE_USER            Delete an existing user.\n"
"    GET_USERS              Get all users.\n"
"    GET_VERSION            Get the OpenVAS Administrator Protocol version.\n"
"    DESCRIBE_FEED          Get a description of the feed this administrator synchronizes with.\n"
"    SYNC_FEED              Synchronize with an NVT feed.\n"
"    HELP                   Get this help text.\n";


/* Status codes. */

/** @todo Share with manager. */

/* HTTP status codes used:
 *
 *     200 OK
 *     201 Created
 *     202 Accepted
 *     400 Bad request
 *     401 Must auth
 *     404 Missing
 *     409 Busy/Conflict
 */

/**
 * @brief Response code for a syntax error.
 */
#define STATUS_ERROR_SYNTAX            "400"

/**
 * @brief Response code when authorisation is required.
 */
#define STATUS_ERROR_MUST_AUTH         "401"

/**
 * @brief Response code when authorisation is required.
 */
#define STATUS_ERROR_MUST_AUTH_TEXT    "Authenticate first"

/**
 * @brief Response code for a missing resource.
 */
#define STATUS_ERROR_MISSING           "404"

/**
 * @brief Response code text for a missing resource.
 */
#define STATUS_ERROR_MISSING_TEXT      "Resource missing"

/**
 * @brief Response code for a busy resource.
 */
#define STATUS_ERROR_BUSY              "409"

/**
 * @brief Response code text for a busy resource.
 */
#define STATUS_ERROR_BUSY_TEXT         "Resource busy"

/**
 * @brief Response code when authorisation failed.
 */
#define STATUS_ERROR_AUTH_FAILED       "400"

/**
 * @brief Response code text when authorisation failed.
 */
#define STATUS_ERROR_AUTH_FAILED_TEXT  "Authentication failed"

/**
 * @brief Response code on success.
 */
#define STATUS_OK                      "200"

/**
 * @brief Response code text on success.
 */
#define STATUS_OK_TEXT                 "OK"

/**
 * @brief Response code on success, when a resource is created.
 */
#define STATUS_OK_CREATED              "201"

/**
 * @brief Response code on success, when a resource is created.
 */
#define STATUS_OK_CREATED_TEXT         "OK, resource created"

/**
 * @brief Response code on success, when the operation will finish later.
 */
#define STATUS_OK_REQUESTED            "202"

/**
 * @brief Response code text on success, when the operation will finish later.
 */
#define STATUS_OK_REQUESTED_TEXT       "OK, request submitted"

/**
 * @brief Response code for an internal error.
 */
#define STATUS_INTERNAL_ERROR          "500"

/**
 * @brief Response code text for an internal error.
 */
#define STATUS_INTERNAL_ERROR_TEXT     "Internal error"

/**
 * @brief Response code when a service is down.
 */
#define STATUS_SERVICE_DOWN            "503"

/**
 * @brief Response code text when a service is down.
 */
#define STATUS_SERVICE_DOWN_TEXT       "Service temporarily down"


/* Global variables. */

/**
 * @brief Generic integer variable for communicating between the callbacks.
 */
int current_int_1;

/**
 * @brief Generic integer variable for communicating between the callbacks.
 */
int current_int_2;

/**
 * @brief Generic string variable for communicating between the callbacks.
 */
static /*@null@*/ /*@only@*/ char*
current_string_1 = NULL;

/**
 * @brief Generic string variable for communicating between the callbacks.
 */
static /*@null@*/ /*@only@*/ char*
current_string_2 = NULL;

/**
 * @brief Generic string variable for communicating between the callbacks.
 */
static /*@null@*/ /*@only@*/ char*
current_string_3 = NULL;

/**
 * @brief Buffer of output to the client.
 */
char to_client[TO_CLIENT_BUFFER_SIZE];

/**
 * @brief The start of the data in the \ref to_client buffer.
 */
buffer_size_t to_client_start = 0;
/**
 * @brief The end of the data in the \ref to_client buffer.
 */
buffer_size_t to_client_end = 0;

/**
 * @brief Client input parsing context.
 */
static /*@null@*/ /*@only@*/ GMarkupParseContext*
xml_context = NULL;

/**
 * @brief Client input parser.
 */
static GMarkupParser xml_parser;


/* Client state. */

/**
 * @brief Possible states of the client.
 */
typedef enum
{
  CLIENT_TOP,
  CLIENT_AUTHENTIC,

  CLIENT_AUTHENTICATE,
  CLIENT_AUTHENTIC_COMMANDS,
  CLIENT_COMMANDS,
  CLIENT_CREATE_USER,
  CLIENT_CREATE_USER_NAME,
  CLIENT_CREATE_USER_PASSWORD,
  CLIENT_CREATE_USER_ROLE,
  CLIENT_CREDENTIALS,
  CLIENT_CREDENTIALS_PASSWORD,
  CLIENT_CREDENTIALS_USERNAME,
  CLIENT_DELETE_USER,
  CLIENT_DESCRIBE_FEED,
  CLIENT_GET_SETTINGS,
  CLIENT_GET_USERS,
  CLIENT_HELP,
  CLIENT_SYNC_FEED,
  CLIENT_VERSION
} client_state_t;

/**
 * @brief The state of the client.
 */
static client_state_t client_state = CLIENT_TOP;

/**
 * @brief Set the client state.
 */
static void
set_client_state (client_state_t state)
{
  client_state = state;
  tracef ("   client state set: %i\n", client_state);
}


/* Communication. */

/**
 * @brief Send a response message to the client.
 *
 * Queue a message in \ref to_client.
 *
 * @param[in]  msg  The message, a string.
 *
 * @return TRUE if out of space in to_client, else FALSE.
 */
static gboolean
send_to_client (char* msg)
{
  assert (to_client_end <= TO_CLIENT_BUFFER_SIZE);
  if (((buffer_size_t) TO_CLIENT_BUFFER_SIZE) - to_client_end
      < strlen (msg))
    {
      tracef ("   send_to_client out of space (%i < %i)\n",
              ((buffer_size_t) TO_CLIENT_BUFFER_SIZE) - to_client_end,
              strlen (msg));
      return TRUE;
    }
  memmove (to_client + to_client_end, msg, strlen (msg));
  tracef ("-> client: %s\n", msg);
  to_client_end += strlen (msg);
  return FALSE;
}

/**
 * @brief Send an XML element error response message to the client.
 *
 * @param[in]  command  Command name.
 * @param[in]  element  Element name.
 *
 * @return TRUE if out of space in to_client, else FALSE.
 */
static gboolean
send_element_error_to_client (const char* command, const char* element)
{
  gchar *msg;
  gboolean ret;

  msg = g_strdup_printf ("<%s_response status=\""
                         STATUS_ERROR_SYNTAX
                         "\" status_text=\"Bogus element: %s\"/>",
                         command,
                         element);
  ret = send_to_client (msg);
  g_free (msg);
  return ret;
}

#if 0
/**
 * @brief Send an XML find error response message to the client.
 *
 * @param[in]  command  Command name.
 * @param[in]  type     Resource type.
 * @param[in]  id       Resource ID.
 *
 * @return TRUE if out of space in to_client, else FALSE.
 */
static gboolean
send_find_error_to_client (const char* command, const char* type,
                           const char* id)
{
  gchar *msg;
  gboolean ret;

  msg = g_strdup_printf ("<%s_response status=\""
                         STATUS_ERROR_MISSING
                         "\" status_text=\"Failed to find %s '%s'\"/>",
                         command, type, id);
  ret = send_to_client (msg);
  g_free (msg);
  return ret;
}
#endif

/**
 * @brief Set an out of space parse error on a GError.
 *
 * @param [out]  error  The error.
 */
static void
error_send_to_client (GError** error)
{
  tracef ("   send_to_client out of space in to_client\n");
  g_set_error (error, G_MARKUP_ERROR, G_MARKUP_ERROR_PARSE,
               "Administrator out of space for reply to client.");
}


/* XML parser handlers. */

/**
 * @brief Expand to XML for a STATUS_ERROR_SYNTAX response.
 *
 * @param  tag   Name of the command generating the response.
 * @param  text  Text for the status_text attribute of the response.
 */
#define XML_ERROR_SYNTAX(tag, text)                      \
 "<" tag "_response"                                     \
 " status=\"" STATUS_ERROR_SYNTAX "\""                   \
 " status_text=\"" text "\"/>"

/**
 * @brief Expand to XML for a STATUS_ERROR_MISSING response.
 *
 * @param  tag  Name of the command generating the response.
 */
#define XML_ERROR_MISSING(tag)                           \
 "<" tag "_response"                                     \
 " status=\"" STATUS_ERROR_MISSING "\""                  \
 " status_text=\"" STATUS_ERROR_MISSING_TEXT "\"/>"

/**
 * @brief Expand to XML for a STATUS_ERROR_AUTH_FAILED response.
 *
 * @param  tag  Name of the command generating the response.
 */
#define XML_ERROR_AUTH_FAILED(tag)                       \
 "<" tag "_response"                                     \
 " status=\"" STATUS_ERROR_AUTH_FAILED "\""              \
 " status_text=\"" STATUS_ERROR_AUTH_FAILED_TEXT "\"/>"

/**
 * @brief Expand to XML for a STATUS_OK response.
 *
 * @param  tag  Name of the command generating the response.
 */
#define XML_OK(tag)                                      \
 "<" tag "_response"                                     \
 " status=\"" STATUS_OK "\""                             \
 " status_text=\"" STATUS_OK_TEXT "\"/>"

/**
 * @brief Expand to XML for a STATUS_OK_CREATED response.
 *
 * @param  tag  Name of the command generating the response.
 */
#define XML_OK_CREATED(tag)                              \
 "<" tag "_response"                                     \
 " status=\"" STATUS_OK_CREATED "\""                     \
 " status_text=\"" STATUS_OK_CREATED_TEXT "\"/>"

/**
 * @brief Expand to XML for a STATUS_OK_REQUESTED response.
 *
 * @param  tag  Name of the command generating the response.
 */
#define XML_OK_REQUESTED(tag)                            \
 "<" tag "_response"                                     \
 " status=\"" STATUS_OK_REQUESTED "\""                   \
 " status_text=\"" STATUS_OK_REQUESTED_TEXT "\"/>"

/**
 * @brief Expand to XML for a STATUS_INTERNAL_ERROR response.
 *
 * @param  tag  Name of the command generating the response.
 */
#define XML_INTERNAL_ERROR(tag)                          \
 "<" tag "_response"                                     \
 " status=\"" STATUS_INTERNAL_ERROR "\""                 \
 " status_text=\"" STATUS_INTERNAL_ERROR_TEXT "\"/>"

/**
 * @brief Expand to XML for a STATUS_SERVICE_DOWN response.
 *
 * @param  tag  Name of the command generating the response.
 */
#define XML_SERVICE_DOWN(tag)                            \
 "<" tag "_response"                                     \
 " status=\"" STATUS_SERVICE_DOWN "\""                   \
 " status_text=\"" STATUS_SERVICE_DOWN_TEXT "\"/>"

/**
 * @brief Find an attribute in a parser callback list of attributes.
 *
 * @param[in]   attribute_names   List of names.
 * @param[in]   attribute_values  List of values.
 * @param[in]   attribute_name    Name of sought attribute.
 * @param[out]  attribute_value   Attribute value return.
 *
 * @return 1 if found, else 0.
 */
int
find_attribute (const gchar **attribute_names,
                const gchar **attribute_values,
                const char *attribute_name,
                const gchar **attribute_value)
{
  while (*attribute_names && *attribute_values)
    if (strcmp (*attribute_names, attribute_name))
      attribute_names++, attribute_values++;
    else
      {
        *attribute_value = *attribute_values;
        return 1;
      }
  return 0;
}

/** @cond STATIC */

/**
 * @brief Send response message to client, returning on fail.
 *
 * Queue a message in \ref to_client with \ref send_to_client.  On failure
 * call \ref error_send_to_client on a GError* called "error" and do a return.
 *
 * @param[in]   msg    The message, a string.
 */
#define SEND_TO_CLIENT_OR_FAIL(msg)                                          \
  do                                                                         \
    {                                                                        \
      if (send_to_client (msg))                                              \
        {                                                                    \
          error_send_to_client (error);                                      \
          return;                                                            \
        }                                                                    \
    }                                                                        \
  while (0)

/**
 * @brief Send response message to client, returning on fail.
 *
 * Queue a message in \ref to_client with \ref send_to_client.  On failure
 * call \ref error_send_to_client on a GError* called "error" and do a return.
 *
 * @param[in]   format    Format string for message.
 * @param[in]   args      Arguments for format string.
 */
#define SENDF_TO_CLIENT_OR_FAIL(format, args...)                             \
  do                                                                         \
    {                                                                        \
      gchar* msg = g_markup_printf_escaped (format , ## args);               \
      if (send_to_client (msg))                                              \
        {                                                                    \
          g_free (msg);                                                      \
          error_send_to_client (error);                                      \
          return;                                                            \
        }                                                                    \
      g_free (msg);                                                          \
    }                                                                        \
  while (0)

/** @endcond */

/**
 * @brief Handle the start of an OAP XML element.
 *
 * React to the start of an XML element according to the current value
 * of \ref client_state, usually adjusting \ref client_state to indicate
 * the change (with \ref set_client_state).  Call \ref send_to_client to
 * queue any responses for the client.
 *
 * Set error parameter on encountering an error.
 *
 * @param[in]  context           Parser context.
 * @param[in]  element_name      XML element name.
 * @param[in]  attribute_names   XML attribute name.
 * @param[in]  attribute_values  XML attribute values.
 * @param[in]  user_data         Dummy parameter.
 * @param[in]  error             Error parameter.
 */
static void
oap_xml_handle_start_element (/*@unused@*/ GMarkupParseContext* context,
                              const gchar *element_name,
                              const gchar **attribute_names,
                              const gchar **attribute_values,
                              /*@unused@*/ gpointer user_data,
                              GError **error)
{
  tracef ("   XML  start: %s (%i)\n", element_name, client_state);

  switch (client_state)
    {
      case CLIENT_TOP:
      case CLIENT_COMMANDS:
        if (strcasecmp ("AUTHENTICATE", element_name) == 0)
          {
            assert (current_credentials.username == NULL);
            assert (current_credentials.password == NULL);
            set_client_state (CLIENT_AUTHENTICATE);
          }
        else if (strcasecmp ("COMMANDS", element_name) == 0)
          {
            SENDF_TO_CLIENT_OR_FAIL
             ("<commands_response"
              " status=\"" STATUS_OK "\" status_text=\"" STATUS_OK_TEXT "\">");
            set_client_state (CLIENT_COMMANDS);
          }
        else
          {
            // TODO: If one of other commands, STATUS_ERROR_MUST_AUTH
            if (send_to_client
                 (XML_ERROR_SYNTAX ("oap",
                                    "First command must be AUTHENTICATE")))
              {
                error_send_to_client (error);
                return;
              }
            g_set_error (error, G_MARKUP_ERROR, G_MARKUP_ERROR_UNKNOWN_ELEMENT,
                         "Must authenticate first.");
          }
        break;

      case CLIENT_AUTHENTIC:
      case CLIENT_AUTHENTIC_COMMANDS:
        if (strcasecmp ("AUTHENTICATE", element_name) == 0)
          {
            /* Any config backend cleanup goes here. */
            free_credentials (&current_credentials);
            set_client_state (CLIENT_AUTHENTICATE);
          }
        else if (strcasecmp ("COMMANDS", element_name) == 0)
          {
            SEND_TO_CLIENT_OR_FAIL
             ("<commands_response"
              " status=\"" STATUS_OK "\" status_text=\"" STATUS_OK_TEXT "\">");
            set_client_state (CLIENT_AUTHENTIC_COMMANDS);
          }
        else if (strcasecmp ("CREATE_USER", element_name) == 0)
          {
            assert (current_string_1 == NULL);
            assert (current_string_2 == NULL);
            assert (current_string_3 == NULL);
            set_client_state (CLIENT_CREATE_USER);
          }
        else if (strcasecmp ("DELETE_USER", element_name) == 0)
          {
            const gchar* attribute;
            if (find_attribute (attribute_names, attribute_values,
                                "name", &attribute))
              openvas_append_string (&current_string_1, attribute);
            set_client_state (CLIENT_DELETE_USER);
          }
        else if (strcasecmp ("GET_USERS", element_name) == 0)
          set_client_state (CLIENT_GET_USERS);
        else if (strcasecmp ("HELP", element_name) == 0)
          set_client_state (CLIENT_HELP);
        else if (strcasecmp ("GET_VERSION", element_name) == 0)
          set_client_state (CLIENT_VERSION);
        else if (strcasecmp ("DESCRIBE_FEED", element_name) == 0)
          set_client_state (CLIENT_DESCRIBE_FEED);
        else if (strcasecmp ("SYNC_FEED", element_name) == 0)
          set_client_state (CLIENT_SYNC_FEED);
        else if (strcasecmp ("GET_SETTINGS", element_name) == 0)
          set_client_state (CLIENT_GET_SETTINGS);
        else
          {
            if (send_to_client (XML_ERROR_SYNTAX ("oap", "Bogus command name")))
              {
                error_send_to_client (error);
                return;
              }
            g_set_error (error,
                         G_MARKUP_ERROR,
                         G_MARKUP_ERROR_UNKNOWN_ELEMENT,
                         "Error");
          }
        break;

      case CLIENT_AUTHENTICATE:
        if (strcasecmp ("CREDENTIALS", element_name) == 0)
          set_client_state (CLIENT_CREDENTIALS);
        else
          {
            if (send_element_error_to_client ("authenticate", element_name))
              {
                error_send_to_client (error);
                return;
              }
            free_credentials (&current_credentials);
            set_client_state (CLIENT_TOP);
            g_set_error (error,
                         G_MARKUP_ERROR,
                         G_MARKUP_ERROR_UNKNOWN_ELEMENT,
                         "Error");
          }
        break;

      case CLIENT_CREDENTIALS:
        if (strcasecmp ("USERNAME", element_name) == 0)
          set_client_state (CLIENT_CREDENTIALS_USERNAME);
        else if (strcasecmp ("PASSWORD", element_name) == 0)
          set_client_state (CLIENT_CREDENTIALS_PASSWORD);
        else
          {
            if (send_element_error_to_client ("authenticate", element_name))
              {
                error_send_to_client (error);
                return;
              }
            free_credentials (&current_credentials);
            set_client_state (CLIENT_TOP);
            g_set_error (error,
                         G_MARKUP_ERROR,
                         G_MARKUP_ERROR_UNKNOWN_ELEMENT,
                         "Error");
          }
        break;

      case CLIENT_DELETE_USER:
        if (send_element_error_to_client ("delete_user", element_name))
          {
            error_send_to_client (error);
            return;
          }
        set_client_state (CLIENT_AUTHENTIC);
        g_set_error (error,
                     G_MARKUP_ERROR,
                     G_MARKUP_ERROR_UNKNOWN_ELEMENT,
                     "Error");
        break;

      case CLIENT_GET_USERS:
          {
            if (send_element_error_to_client ("get_users", element_name))
              {
                error_send_to_client (error);
                return;
              }
            set_client_state (CLIENT_AUTHENTIC);
            g_set_error (error,
                         G_MARKUP_ERROR,
                         G_MARKUP_ERROR_UNKNOWN_ELEMENT,
                         "Error");
          }
        break;

      case CLIENT_HELP:
        {
          if (send_element_error_to_client ("help", element_name))
            {
              error_send_to_client (error);
              return;
            }
          set_client_state (CLIENT_AUTHENTIC);
          g_set_error (error,
                       G_MARKUP_ERROR,
                       G_MARKUP_ERROR_UNKNOWN_ELEMENT,
                       "Error");
        }
        break;

      case CLIENT_CREATE_USER:
        if (strcasecmp ("NAME", element_name) == 0)
          {
            assert (current_string_1 == NULL);
            set_client_state (CLIENT_CREATE_USER_NAME);
          }
        else if (strcasecmp ("PASSWORD", element_name) == 0)
          {
            assert (current_string_2 == NULL);
            set_client_state (CLIENT_CREATE_USER_PASSWORD);
          }
        else if (strcasecmp ("ROLE", element_name) == 0)
          {
            assert (current_string_3 == NULL);
            set_client_state (CLIENT_CREATE_USER_ROLE);
          }
        else
          {
            if (send_element_error_to_client ("create_user", element_name))
              {
                error_send_to_client (error);
                return;
              }
            set_client_state (CLIENT_AUTHENTIC);
            g_set_error (error,
                         G_MARKUP_ERROR,
                         G_MARKUP_ERROR_UNKNOWN_ELEMENT,
                         "Error");
          }
        break;

      default:
        assert (0);
        // FIX respond fail to client
        g_set_error (error,
                     G_MARKUP_ERROR,
                     G_MARKUP_ERROR_PARSE,
                     "Administrator programming error.");
        break;
    }

  return;
}

/**
 * @brief Handle the end of an OAP XML element.
 *
 * React to the end of an XML element according to the current value
 * of \ref client_state, usually adjusting \ref client_state to indicate
 * the change (with \ref set_client_state).  Call \ref send_to_client to queue
 * any responses for the client.  Call the task utilities to adjust the
 * tasks (for example \ref start_task, \ref stop_task, \ref set_task_parameter,
 * \ref delete_task and \ref find_task).
 *
 * Set error parameter on encountering an error.
 *
 * @param[in]  context           Parser context.
 * @param[in]  element_name      XML element name.
 * @param[in]  user_data         Dummy parameter.
 * @param[in]  error             Error parameter.
 */
static void
oap_xml_handle_end_element (/*@unused@*/ GMarkupParseContext* context,
                            const gchar *element_name,
                            /*@unused@*/ gpointer user_data,
                            GError **error)
{
  tracef ("   XML    end: %s\n", element_name);
  switch (client_state)
    {
      case CLIENT_TOP:
        assert (0);
        break;

      case CLIENT_AUTHENTICATE:
        switch (authenticate (&current_credentials))
          {
            case 0:
              /* Any config backend initialisation goes here. */
              SEND_TO_CLIENT_OR_FAIL (XML_OK ("authenticate"));
              set_client_state (CLIENT_AUTHENTIC);
              break;
            case 1:
              free_credentials (&current_credentials);
              SEND_TO_CLIENT_OR_FAIL (XML_ERROR_AUTH_FAILED ("authenticate"));
              set_client_state (CLIENT_TOP);
              break;
            case -1:
            default:
              free_credentials (&current_credentials);
              SEND_TO_CLIENT_OR_FAIL (XML_INTERNAL_ERROR ("authenticate"));
              set_client_state (CLIENT_TOP);
              break;
          }
        break;

      case CLIENT_AUTHENTIC:
      case CLIENT_COMMANDS:
      case CLIENT_AUTHENTIC_COMMANDS:
        assert (strcasecmp ("COMMANDS", element_name) == 0);
        SENDF_TO_CLIENT_OR_FAIL ("</commands_response>");
        break;

      case CLIENT_CREDENTIALS:
        assert (strcasecmp ("CREDENTIALS", element_name) == 0);
        set_client_state (CLIENT_AUTHENTICATE);
        break;

      case CLIENT_CREDENTIALS_USERNAME:
        assert (strcasecmp ("USERNAME", element_name) == 0);
        set_client_state (CLIENT_CREDENTIALS);
        break;

      case CLIENT_CREDENTIALS_PASSWORD:
        assert (strcasecmp ("PASSWORD", element_name) == 0);
        set_client_state (CLIENT_CREDENTIALS);
        break;

      case CLIENT_VERSION:
        SEND_TO_CLIENT_OR_FAIL ("<get_version_response"
                                " status=\"" STATUS_OK "\""
                                " status_text=\"" STATUS_OK_TEXT "\">"
                                "<version preferred=\"yes\">1.0</version>"
                                "</get_version_response>");
        set_client_state (CLIENT_AUTHENTIC);
        break;

      case CLIENT_DELETE_USER:
        if (current_string_1)
          {
            assert (users_dir);
            if (strcmp (current_credentials.username, current_string_1) == 0)
              SEND_TO_CLIENT_OR_FAIL
               (XML_ERROR_SYNTAX ("delete_user",
                                  "Attempt to delete current user"));
            else
              switch (openvas_admin_remove_user (current_string_1, users_dir))
                {
                  case 0:
                    SEND_TO_CLIENT_OR_FAIL (XML_OK ("delete_user"));
                    break;
                  case -2:
                    SEND_TO_CLIENT_OR_FAIL
                     (XML_ERROR_SYNTAX ("delete_user",
                                        "Failed to find user"));
                    break;
                  default:
                  case -1:
                    SEND_TO_CLIENT_OR_FAIL (XML_INTERNAL_ERROR ("delete_user"));
                    break;
                }
            openvas_free_string_var (&current_string_1);
          }
        else
          SEND_TO_CLIENT_OR_FAIL
           (XML_ERROR_SYNTAX ("delete_user",
                              "DELETE_USER requires a name attribute"));
        set_client_state (CLIENT_AUTHENTIC);
        break;

      case CLIENT_HELP:
        SEND_TO_CLIENT_OR_FAIL ("<help_response"
                                " status=\"" STATUS_OK "\""
                                " status_text=\"" STATUS_OK_TEXT "\">");
        SEND_TO_CLIENT_OR_FAIL (help_text);
        SEND_TO_CLIENT_OR_FAIL ("</help_response>");
        set_client_state (CLIENT_AUTHENTIC);
        break;

      case CLIENT_DESCRIBE_FEED:
          {
            gchar *feed_description = NULL;
            gchar *feed_identification = NULL;
            if (openvas_get_sync_script_description (sync_script,
                                                     &feed_description) &&
                openvas_get_sync_script_identification (sync_script,
                                                        &feed_identification))
              {
                gchar **ident = g_strsplit (feed_identification, "|", 6);
                SEND_TO_CLIENT_OR_FAIL ("<describe_feed_response"
                                        " status=\"" STATUS_OK "\""
                                        " status_text=\"" STATUS_OK_TEXT "\">");
                SEND_TO_CLIENT_OR_FAIL ("<feed>");
                SENDF_TO_CLIENT_OR_FAIL ("<name>%s</name>", ident[3]);
                SENDF_TO_CLIENT_OR_FAIL ("<description>%s</description>", feed_description);
                if (openvasad_currently_syncing ())
                  {
                    SEND_TO_CLIENT_OR_FAIL ("<currently_syncing />");
                  }
                SEND_TO_CLIENT_OR_FAIL ("</feed>");
                SEND_TO_CLIENT_OR_FAIL ("</describe_feed_response>");
                g_strfreev (ident);
                g_free (feed_description);
                g_free (feed_identification);
              }
            else
              {
                SEND_TO_CLIENT_OR_FAIL ("<describe_feed_response"
                                        " status=\"" STATUS_OK "\""
                                        " status_text=\"" STATUS_OK_TEXT "\">");
                SEND_TO_CLIENT_OR_FAIL ("<feed>");
                SEND_TO_CLIENT_OR_FAIL ("<name></name>");
                SEND_TO_CLIENT_OR_FAIL ("<description></description>");
                SEND_TO_CLIENT_OR_FAIL ("</feed>");
                SEND_TO_CLIENT_OR_FAIL ("</describe_feed_response>");
              }
            set_client_state (CLIENT_AUTHENTIC);
            break;
          }

      case CLIENT_SYNC_FEED:
          {
            if (openvasad_currently_syncing ())
              {
                SEND_TO_CLIENT_OR_FAIL ("<sync_feed_response"
                                        " status=\"" STATUS_ERROR_BUSY "\""
                                        " status_text=\"" STATUS_ERROR_BUSY_TEXT "\">");
                SEND_TO_CLIENT_OR_FAIL ("</sync_feed_response>");
              }
            openvasad_create_lockfile (sync_script);
            if (openvas_sync_feed (sync_script))
              {
                SEND_TO_CLIENT_OR_FAIL ("<sync_feed_response"
                                        " status=\"" STATUS_OK "\""
                                        " status_text=\"" STATUS_OK_TEXT "\">");
                SEND_TO_CLIENT_OR_FAIL ("</sync_feed_response>");
                set_client_state (CLIENT_AUTHENTIC);
              }
            else
              {
                SEND_TO_CLIENT_OR_FAIL (XML_INTERNAL_ERROR ("sync_feed"));
              }
            openvasad_remove_lockfile ();
            set_client_state (CLIENT_AUTHENTIC);
            break;
          }

      case CLIENT_GET_SETTINGS:
          {
            GHashTable *scanner_settings_hash_table = get_all_settings
              (scanner_config_file, "Misc");

            if (scanner_settings_hash_table != NULL)
              {
                GString *scanner_settings = g_string_new ("");
                g_hash_table_foreach (scanner_settings_hash_table,
                                      build_settings_xml, scanner_settings);
                SEND_TO_CLIENT_OR_FAIL ("<get_settings_response"
                                        " status=\"" STATUS_OK "\""
                                        " status_text=\"" STATUS_OK_TEXT "\">");
                SENDF_TO_CLIENT_OR_FAIL ("<scanner_settings sourcefile=\"%s\">",
                                         scanner_config_file);
                SEND_TO_CLIENT_OR_FAIL (scanner_settings->str);
                SEND_TO_CLIENT_OR_FAIL ("</scanner_settings>");

                SEND_TO_CLIENT_OR_FAIL ("</get_settings_response>");
                g_string_free (scanner_settings, TRUE);
                g_hash_table_destroy (scanner_settings_hash_table);
              }
            else
              {
                SENDF_TO_CLIENT_OR_FAIL ("<get_settings_response"
                                        " status=\"" STATUS_ERROR_MISSING "\""
                                        " status_text=\"Could not read settings from %s\">",
                                        scanner_config_file);
                SEND_TO_CLIENT_OR_FAIL ("</get_settings_response>");
              }
            set_client_state (CLIENT_AUTHENTIC);
            break;
          }

      case CLIENT_CREATE_USER:
        {
          assert (strcasecmp ("CREATE_USER", element_name) == 0);
          assert (users_dir);

          if (current_string_1 == NULL || strlen (current_string_1) == 0)
            SEND_TO_CLIENT_OR_FAIL
             (XML_ERROR_SYNTAX ("create_user",
                                "CREATE_USER requires a name"));
          else
            switch (openvas_admin_add_user
                     (current_string_1,           /* Name. */
                      current_string_2 ?: "",     /* Pass. */
                      current_string_3 ?: "User", /* Role. */
                      users_dir))
              {
                case 0:
                  SEND_TO_CLIENT_OR_FAIL (XML_OK ("create_user"));
                  break;
                case -2:
                  SEND_TO_CLIENT_OR_FAIL
                   (XML_ERROR_SYNTAX ("create_user", "User already exists"));
                  break;
                case -1:
                default:
                  SEND_TO_CLIENT_OR_FAIL (XML_INTERNAL_ERROR ("create_user"));
                  break;
              }
          openvas_free_string_var (&current_string_1);
          openvas_free_string_var (&current_string_2);
          openvas_free_string_var (&current_string_3);
          set_client_state (CLIENT_AUTHENTIC);
          break;
        }
      case CLIENT_CREATE_USER_NAME:
        assert (strcasecmp ("NAME", element_name) == 0);
        set_client_state (CLIENT_CREATE_USER);
        break;
      case CLIENT_CREATE_USER_PASSWORD:
        assert (strcasecmp ("PASSWORD", element_name) == 0);
        set_client_state (CLIENT_CREATE_USER);
        break;
      case CLIENT_CREATE_USER_ROLE:
        assert (strcasecmp ("ROLE", element_name) == 0);
        set_client_state (CLIENT_CREATE_USER);
        break;

      case CLIENT_GET_USERS:
        {
          GSList *users, *user;
          assert (strcasecmp ("GET_USERS", element_name) == 0);
          SEND_TO_CLIENT_OR_FAIL ("<get_users_response"
                                  " status=\"" STATUS_OK "\""
                                  " status_text=\"" STATUS_OK_TEXT "\">");
          user = users = openvas_admin_list_users (users_dir);
          while (user)
            {
              SENDF_TO_CLIENT_OR_FAIL ("<user>"
                                       "<name>%s</name>"
                                       "<role>%s</role>"
                                       "</user>",
                                       (gchar*) user->data,
                                       openvas_is_user_admin (user->data)
                                       ? "Administrator" : "User");
              g_free (user->data);
              user = g_slist_next (user);
            }
          g_slist_free (users);
          SEND_TO_CLIENT_OR_FAIL ("</get_users_response>");
          set_client_state (CLIENT_AUTHENTIC);
          break;
        }

      default:
        assert (0);
        break;
    }
}

/**
 * @brief Handle the addition of text to an OAP XML element.
 *
 * React to the addition of text to the value of an XML element.
 * React according to the current value of \ref client_state,
 * usually appending the text to some part of the current task
 * (\ref current_client_task) with functions like \ref openvas_append_text,
 * \ref add_task_description_line and \ref append_to_task_comment.
 *
 * @param[in]  context           Parser context.
 * @param[in]  text              The text.
 * @param[in]  text_len          Length of the text.
 * @param[in]  user_data         Dummy parameter.
 * @param[in]  error             Error parameter.
 */
static void
oap_xml_handle_text (/*@unused@*/ GMarkupParseContext* context,
                     const gchar *text,
                     gsize text_len,
                     /*@unused@*/ gpointer user_data,
                     /*@unused@*/ GError **error)
{
  if (text_len == 0) return;
  tracef ("   XML   text: %s\n", text);
  switch (client_state)
    {
      case CLIENT_CREDENTIALS_USERNAME:
        append_to_credentials_username (&current_credentials, text, text_len);
        break;
      case CLIENT_CREDENTIALS_PASSWORD:
        append_to_credentials_password (&current_credentials, text, text_len);
        break;

      case CLIENT_CREATE_USER_NAME:
        openvas_append_text (&current_string_1, text, text_len);
        break;
      case CLIENT_CREATE_USER_PASSWORD:
        openvas_append_text (&current_string_2, text, text_len);
        break;
      case CLIENT_CREATE_USER_ROLE:
        openvas_append_text (&current_string_3, text, text_len);
        break;

      default:
        /* Just pass over the text. */
        break;
    }
}

/**
 * @brief Handle an OAP XML parsing error.
 *
 * Simply leave the error for the caller of the parser to handle.
 *
 * @param[in]  context           Parser context.
 * @param[in]  error             The error.
 * @param[in]  user_data         Dummy parameter.
 */
static void
oap_xml_handle_error (/*@unused@*/ GMarkupParseContext* context,
                      GError *error,
                      /*@unused@*/ gpointer user_data)
{
  tracef ("   XML ERROR %s\n", error->message);
}


/* OAP input processor. */

// FIX probably should pass to process_oap_client_input
extern char from_client[];
extern buffer_size_t from_client_start;
extern buffer_size_t from_client_end;

/**
 * @brief Initialise OAP library.
 *
 * @param[in]  log_config       Logging configuration list.
 * @param[in]  users_directory  Directory containing user info.
 * @param[in]  synchronization_script  The script to use for feed
 * synchronization.
 * @param[in]  configuration_file  Scanner Configuration file.
 *
 * @return 0 success, -1 error.
 */
int
init_oap (GSList *log_config, const gchar *users_directory,
          const gchar *synchronization_script,
          const gchar *configuration_file)
{
  if (users_directory == NULL) return -1;
  if (synchronization_script == NULL) return -1;
  if (configuration_file == NULL) return -1;
  g_log_set_handler (G_LOG_DOMAIN,
                     ALL_LOG_LEVELS,
                     (GLogFunc) openvas_log_func,
                     log_config);
  users_dir = users_directory;
  sync_script = synchronization_script;
  scanner_config_file = configuration_file;
  current_credentials.username = NULL;
  current_credentials.password = NULL;
  return 0;
}

/**
 * @brief Initialise OAP library data for a process.
 *
 * This should run once per process, before the first call to \ref
 * process_oap_client_input.
 */
void
init_oap_process ()
{
  /* Create the XML parser. */
  xml_parser.start_element = oap_xml_handle_start_element;
  xml_parser.end_element = oap_xml_handle_end_element;
  xml_parser.text = oap_xml_handle_text;
  xml_parser.passthrough = NULL;
  xml_parser.error = oap_xml_handle_error;
  if (xml_context) g_free (xml_context);
  xml_context = g_markup_parse_context_new (&xml_parser,
                                            0,
                                            NULL,
                                            NULL);
}

/**
 * @brief Process any XML available in \ref from_client.
 *
 * \if STATIC
 *
 * Call the XML parser and let the callback functions do the work
 * (\ref oap_xml_handle_start_element, \ref oap_xml_handle_end_element,
 * \ref oap_xml_handle_text and \ref oap_xml_handle_error).
 *
 * The callback functions will queue any replies for the client in
 * \ref to_client (using \ref send_to_client).
 *
 * \endif
 *
 * @return 0 success, -1 error, -2 too little space in \ref to_client,
 *         -4 XML syntax error.
 */
int
process_oap_client_input ()
{
  gboolean success;
  GError* error = NULL;

  if (xml_context == NULL) return -1;

  success = g_markup_parse_context_parse (xml_context,
                                          from_client + from_client_start,
                                          from_client_end - from_client_start,
                                          &error);
  if (success == FALSE)
    {
      int err;
      if (error)
        {
          err = -4;
          if (g_error_matches (error,
                               G_MARKUP_ERROR,
                               G_MARKUP_ERROR_UNKNOWN_ELEMENT))
            tracef ("   client error: G_MARKUP_ERROR_UNKNOWN_ELEMENT\n");
          else if (g_error_matches (error,
                                    G_MARKUP_ERROR,
                                    G_MARKUP_ERROR_INVALID_CONTENT))
            tracef ("   client error: G_MARKUP_ERROR_INVALID_CONTENT\n");
          else if (g_error_matches (error,
                                    G_MARKUP_ERROR,
                                    G_MARKUP_ERROR_UNKNOWN_ATTRIBUTE))
            tracef ("   client error: G_MARKUP_ERROR_UNKNOWN_ATTRIBUTE\n");
          else
            err = -1;
          g_message ("   Failed to parse client XML: %s\n", error->message);
          g_error_free (error);
        }
      else
        err = -1;
      /* In all error cases the caller must cease to call this function as it
       * would be too hard, if possible at all, to figure out the position of
       * start of the next command. */
      return err;
    }
  from_client_end = from_client_start = 0;
  return 0;
}
